@@ -1,13 +1,8 @@
 /* main.c */
 
-/*  $RCSfile: main.c,v $
- *  $Revision: 14020.15 $
- *  $Date: 93/07/09 11:50:12 $
- */
-
 #define _main_c_
 
-#define FTP_VERSION "1.7.1 (March 27, 1994)"
+#define FTP_VERSION "1.9.5 (October 29, 1995)"
 
 /* #define BETA 1 */ /* If defined, it prints a little warning message. */
 
@@ -34,6 +29,10 @@
 #	endif
 #endif	/* CURSES */
 
+#if defined(CURSES) && defined(SGTTYB)
+#       include <sgtty.h>
+#endif
+
 #include "util.h"
 #include "cmds.h"
 #include "main.h"
@@ -51,7 +50,9 @@ int					toatty;				/* output is to a terminal */
 int					doing_script;		/* is a file being <redirected to me? */
 char				*altarg;			/* argv[1] with no shell-like preprocessing  */
 struct servent		serv;				/* service spec for tcp/ftp */
+static char			pad2a[8] = "Pad 2a";	/* SunOS overwrites jmp_bufs... */
 jmp_buf				toplevel;			/* non-local goto stuff for cmd scanner */
+static char			pad2b[8] = "Pad 2b";
 char				*line;				/* input line buffer */
 char				*stringbase;		/* current scan point in line buffer */
 char				*argbuf;			/* argument storage buffer */
@@ -88,6 +89,15 @@ char				*tcap_reverse = "\033&dB";
 
 #else
 
+#ifdef NO_FORMATTING
+
+char                            *tcap_normal = "";
+char                            *tcap_boldface = "";
+char                            *tcap_underline = "";
+char                            *tcap_reverse = "";
+
+#else
+
 char                            *tcap_normal = "\033[0m";       /* Default ANSI escapes */
 char                            *tcap_boldface = "\033[1m";
 char                            *tcap_underline = "\033[4m";
@@ -95,6 +105,8 @@ char                            *tcap_reverse = "\033[7m";
 
 #endif
 
+#endif
+
 size_t				tcl_normal = 4,		/* lengths of the above strings. */
 					tcl_bold = 4,
 					tcl_uline = 4,
@@ -105,7 +117,7 @@ static char			tcbuf[2048];
 #endif
 
 /* main.c externs */
-extern int			debug, verbose, mprompt;
+extern int			debug, verbose, mprompt, passivemode;
 extern int			options, cpend, data, connected, logged_in;
 extern int			curtype, macnum, remote_is_unix;
 extern FILE			*cout;
@@ -164,6 +176,7 @@ Re-compile, this time with -DZCAT=\\\"/path/to/zcat\\\".\n");
 	mprompt = dMPROMPT;
 	debug = dDEBUG;
 	verbose = dVERBOSE;
+	passivemode = dPASSIVE;
 	(void) Strncpy(vstr, short_verbose_msgs[verbose+1]);
 
 	(void) Strncpy(curtypename, dTYPESTR);
@@ -223,7 +236,7 @@ Re-compile, this time with -DZCAT=\\\"/path/to/zcat\\\".\n");
 
 	ignore_rc = 0;
 	(void) strcpy(oline, "open ");
-	while ((opt = Getopt(argc, argv, "D:V:INRHaicmup:rd:g:")) >= 0) {
+	while ((opt = Getopt(argc, argv, "D:V:INPRHaicmup:rd:g:")) >= 0) {
 		switch(opt) {
 			case 'a':
 			case 'c':
@@ -259,6 +272,10 @@ Re-compile, this time with -DZCAT=\\\"/path/to/zcat\\\".\n");
 				++ignore_rc;
 				break;
 
+			case 'P':
+				passivemode = !passivemode;
+				break;
+
 			case 'H':
 				(void) show_version(0, NULL);
 				exit (0);
@@ -271,6 +288,7 @@ Program Options:\n\
     -H     : Show version and compilation information.\n\
     -I     : Toggle interactive (mprompt) mode.\n\
     -N     : Toggle reading of the .netrc/.ncftprc.\n\
+    -P     : Toggle passive mode ftp (for use behind firewalls).\n\
     -V x   : Set verbosity to level x (-1,0,1,2).\n\
 Open Options:\n\
     -a     : Open anonymously (this is the default).\n\
@@ -331,7 +349,9 @@ Examples:\n\
 		makeargv();
 		/* setpeer uses this to tell if it was called from the cmd-line. */
 		eventnumber = 0L;
-		(void) cmdOpen(margc, margv);
+		if (cmdOpen(margc, margv) != NOERR) {
+			exit(1);
+		}
 	}
 	eventnumber = 1L;
 
@@ -379,7 +399,8 @@ For testing purposes only.  Do not re-distribute or subject to novice users."
 		(void) Signal(SIGPIPE, lostpeer);
 	}
 	for (;;) {
-		(void) cmdscanner(top);
+		if (cmdscanner(top) && !fromatty)
+			exit(1);
 		top = 1;
 	}
 }	/* main */
@@ -406,8 +427,21 @@ int getuserinfo(void)
 	home = uinfo.homedir;	/* for glob.c */
 	pw = NULL;
 #ifdef USE_GETPWUID
+	/* Try to use getpwuid(), but if we have to, fall back to getpwnam(). */
 	pw = getpwuid(getuid());
+	if (pw == NULL) {
+		/* Oh well, try getpwnam() then. */
+		cp = getlogin();
+		if (cp == NULL) {
+			cp = getenv("LOGNAME");
+			if (cp == NULL)
+				cp = getenv("USER");
+		}
+		if (cp != NULL)
+			pw = getpwnam(cp);
+	}
 #else
+	/* Try to use getpwnam(), but if we have to, fall back to getpwuid(). */
 	cp = getlogin();
 	if (cp == NULL) {
 		cp = getenv("LOGNAME");
@@ -416,6 +450,10 @@ int getuserinfo(void)
 	}
 	if (cp != NULL)
 		pw = getpwnam(cp);
+	if (pw == NULL) {
+		/* Oh well, try getpwuid() then. */
+		pw = getpwuid(getuid());
+	}
 #endif
 	if (pw != NULL) {
 		uinfo.uid = pw->pw_uid;
@@ -426,23 +464,32 @@ int getuserinfo(void)
 		else
 			(void) Strncpy(uinfo.homedir, pw->pw_dir);
 		cp = getenv("MAIL");
-		if (cp == NULL)
-			cp = getenv("mail");
-		if (cp == NULL)
-			(void) sprintf(str, "/usr/spool/mail/%s", uinfo.username);
-		else
+		if (cp != NULL) {
 			(void) Strncpy(str, cp);
+		} else {
+			/* Check a few typical mail directories.
+			 * If we don't find it, too bad.  Checking for new mail
+			 * isn't very important anyway.
+			 */
+			(void) sprintf(str, "/usr/spool/mail/%s", uinfo.username);
+			if (access(str, 0) < 0) {
+				(void) sprintf(str, "/var/mail/%s", uinfo.username);
+			}
+		}
 		cp = str;
-
-		/*
-		 * mbox variable may be like MAIL=(28 /usr/mail/me /usr/mail/you),
-		 * so try to find the first mail path.
-		 */
-		while ((*cp != '/') && (*cp != 0))
-			cp++;
-		(void) Strncpy(mail_path, cp);
-		if ((cp = index(mail_path, ' ')) != NULL)
-			*cp = '\0';
+		if (access(cp, 0) < 0) {
+			mail_path[0] = 0;
+		} else {
+			/*
+			 * mbox variable may be like MAIL=(28 /usr/mail/me /usr/mail/you),
+			 * so try to find the first mail path.
+			 */
+			while ((*cp != '/') && (*cp != 0))
+				cp++;
+			(void) Strncpy(mail_path, cp);
+			if ((cp = index(mail_path, ' ')) != NULL)
+				*cp = '\0';
+		}
 		return (0);
 	} else {
 		PERROR("getuserinfo", "Could not get your passwd entry!");
@@ -541,9 +588,10 @@ void lostpeer SIG_PARAMS
 /*
  * Command parser.
  */
-void cmdscanner(int top)
+int cmdscanner(int top)
 {
 	register struct cmd *c;
+	int cmd_status, rcode = 0;
 
 	if (!top)
 		(void) putchar('\n');
@@ -573,13 +621,17 @@ void cmdscanner(int top)
 			(void) printf ("Not connected.\n");
 			continue;
 		}
-		if ((*c->c_handler)(margc, margv) == USAGE)
+		cmd_status = (*c->c_handler)(margc, margv);
+	    if (cmd_status == USAGE)
 			cmd_usage(c);
+	    else if (cmd_status == CMDERR)
+			rcode = 1;
 		if (c->c_handler != help)
 			break;
 	}
 	(void) Signal(SIGINT, intr);
 	(void) Signal(SIGPIPE, lostpeer);
+	return rcode;
 }	/* cmdscanner */
 
 
@@ -1056,6 +1108,10 @@ int termcap_get(char **dest, char *attr)
 void termcap_init(void)
 {
 	char *term;
+#ifdef  SGTTYB
+	struct sgttyb ttyb;
+	extern short ospeed;
+#endif
 
 	if ((term = getenv("TERM")) == NULL) {
 		term = "dumb";  /* TAR */
@@ -1067,7 +1123,7 @@ void termcap_init(void)
 		(void) termcap_get(&tcap_normal, "meuese");
 		if (termcap_get(&tcap_boldface, "md") < 0) {
 			/* Dim-mode is better than nothing... */
-			(void) termcap_get(&tcap_normal, "mh");
+			(void) termcap_get(&tcap_boldface, "mh");
 		}
 		(void) termcap_get(&tcap_underline, "us");
 		(void) termcap_get(&tcap_reverse, "so");
@@ -1075,6 +1131,10 @@ void termcap_init(void)
 		tcl_bold = strlen(tcap_boldface);
 		tcl_uline = strlen(tcap_underline);
 		tcl_rev = strlen(tcap_reverse);
+#ifdef  SGTTYB
+		if (ioctl(fileno(stdout), TIOCGETP, &ttyb) == 0)
+			ospeed = ttyb.sg_ospeed;
+#endif
 	}
 
 }	/* termcap_init */
